/*
    Copyright 2007,2008,2009,2010 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>
#include <sys/stat.h>
#include <openssl/aes.h>

typedef uint8_t     u8;

#ifdef WIN32
    #include <windows.h>

    int camfrog_enum(u8 *ret_clsid);
#else
    #define camfrog_enum(X) -1
    #define stristr strcasestr
    #define stricmp strcasecmp
#endif



#define VER         "0.3.2"
#define BUFFSZ      4096
#define KEYPATH     "Software\\Camfrog"
#define CLSIDKEY    "Software\\Microsoft\\Windows\\CurrentVersion\\Telephony"
#define INTV1       "Internal Value:"
#define INTV2       "I\0n\0t\0e\0r\0n\0a\0l\0 \0V\0a\0l\0u\0e\0:\0"



u8 *get_phclsid_from_file(u8 *fname);
void show_pwd(u8 *msg, u8 *pwd, u8 *clsid);
void camfrog_file(FILE *fd, u8 *clsid);
int hex2byte(u8 *hex);
u8 *camfrog_pwd(u8 *pwd, u8 *clsid);
u8 *camfrogbot_pwd(u8 *pwd, u8 *clsid);



int main(int argc, char *argv[]) {
    FILE    *fd;
    u8      *pwd,
            *clsid  = NULL;

    setbuf(stdout, NULL);

    fputs("\n"
        "CamFrog passwords decrypter "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s [encoded_data/file] [PhDeviceCLSID/file]\n"
            "\n"
            "  this tool can be used to decrypt any encrypted value located in the Camfrog\n"
            "  registry keys like Serial, ActivationData and any password and in any other\n"
            "  file like cf_server.conf, cf_room.conf, settings.xml and so on in %%APPDATA%%\n"
            "\n"
            "  PhDeviceCLSID is the value located in the following registry key:\n"
            "    HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Telephony\\PhDeviceCLSID\n"
            "  this value is required if you want to decrypt a custom ActivationData key\n"
            "  (Camfrog Bot needs this field for both serial and passwords)\n"
            "  note that in Camfrog Bot the PhDeviceCLSID is saved also inside the same\n"
            "  executable or in mdlg.dll near the \"Internal Value:\" string\n"
            "\n", argv[0]);

#ifdef WIN32
        printf("- start registry scanning:\n\n");
        if(!camfrog_enum(NULL)) {
            printf("- no Serial, ActivationData or passwords found in the registry\n\n");
        }
#endif
    } else {
        pwd = strdup(argv[1]);
        if(argc > 2) {
            printf("- check if the second argument is a file\n");
            clsid = get_phclsid_from_file(argv[2]);
            if(!clsid) clsid = strdup(argv[2]);
            printf("- use custom clsid: %s\n", clsid);
            printf("\n");
        }

        fd = fopen(pwd, "rb");
        if(fd) {
            camfrog_file(fd, clsid);
            fclose(fd);
        } else {
            show_pwd("decrypted data", pwd, clsid);
        }
    }

#ifdef WIN32
    printf("\nPress RETURN to quit\n");
    fgetc(stdin);
#endif
    return(0);
}



void myunicode2char(u8 *s) {
    u8      *p;

    for(p = s; *p; p += 2) {
        *s++ = *p;
    }
    *s = 0;
}

u8 *fdload(u8 *fname, int *fsize) {
    struct stat xstat;
    FILE    *fd;
    int     size;
    u8      *buff;

    fd = fopen(fname, "rb");
    if(!fd) return(NULL);
    fstat(fileno(fd), &xstat);
    size = xstat.st_size;
    buff = malloc(size);
    if(!buff) return(NULL);
    fread(buff, 1, size, fd);
    fclose(fd);
    if(fsize) *fsize = size;
    return(buff);
}

// from NetBSD
void *mymemmem(const void *b1, const void *b2, size_t len1, size_t len2) {
    unsigned char *sp  = (unsigned char *) b1;
    unsigned char *pp  = (unsigned char *) b2;
    unsigned char *eos = sp + len1 - len2;

    if(!(b1 && b2 && len1 && len2))
        return NULL;

    while (sp <= eos) {
        if (*sp == *pp)
            if (memcmp(sp, pp, len2) == 0)
                return sp;
        sp++;
    }
    return NULL;
}

u8 *get_phclsid_from_file(u8 *fname) {
    int     tmp;
    u8      *p,
            *clsid;

    if(!fname || !fname[0]) return(NULL);
    p = fdload(fname, &tmp);
    if(!p) return(NULL);
    clsid = mymemmem(p, INTV1, tmp, sizeof(INTV1)-1);
    if(clsid) { // ascii
        clsid += sizeof(INTV1)-1;
    } else {
        clsid = mymemmem(p, INTV2, tmp, sizeof(INTV2)-1);
        if(clsid) { // unicode
            clsid += sizeof(INTV2)-1;
        } else {
            printf("\nError: no \"Internal Value\" found in the file %s\n", fname);
            exit(1);
        }
        myunicode2char(clsid);
    }
    clsid = strdup(clsid);
    free(p);
    return(clsid);
}



#ifdef WIN32
int regkey(HKEY hKey, LPCTSTR lpSubKey, LPTSTR lpValueName, u8 *buff, int len) {
    HKEY    key;

    buff[0] = 0;
    if(RegOpenKeyEx(hKey, lpSubKey, 0, KEY_READ, &key) != ERROR_SUCCESS) {
        return(-1);
    }
    if(RegQueryValueEx(key, lpValueName, NULL, NULL, buff, (void *)&len) != ERROR_SUCCESS) {
        RegCloseKey(key);
        return(-1);
    }
    RegCloseKey(key);
    return(0);
}
int camfrog_enum(u8 *ret_clsid) {
    static u8   clsidbuff[BUFFSZ + 1] = "";
    HKEY    key;
    int     len,
            k,
            tot;
    u8      regbuff[BUFFSZ + 1],
            buff[BUFFSZ + 1],
            *p,
            *c;

    if(!clsidbuff[0]) {
        if(regkey(HKEY_CURRENT_USER, CLSIDKEY, "PhDeviceCLSID", clsidbuff, sizeof(clsidbuff)) < 0) {
            if(regkey(HKEY_LOCAL_MACHINE, CLSIDKEY, "PhDeviceCLSID", clsidbuff, sizeof(clsidbuff)) < 0) {
                // well, it's enough chaotic anyway I try to give priority to CamfrogBot (ret_clsid)
                // those of the server are only for a possible future so are not needed now
                if(ret_clsid && (regkey(HKEY_CURRENT_USER, KEYPATH, "Last Install Dir Bot 5.0", buff, sizeof(buff)) >= 0)) {
                    p = buff + strlen(buff);
                    strcpy(p, "\\CamfrogBot.exe");
                } else if(regkey(HKEY_CURRENT_USER, KEYPATH, "Last Install Dir Camfrog Client", buff, sizeof(buff)) >= 0) {
                    p = buff + strlen(buff);
                    strcpy(p, "\\mdlg.dll");
                } else if(regkey(HKEY_CURRENT_USER, KEYPATH, "Last Install Dir Camfrog Server", buff, sizeof(buff)) >= 0) {
                    p = buff + strlen(buff);
                    strcpy(p, "\\CamfrogServer.exe");
                } else if(regkey(HKEY_CURRENT_USER, KEYPATH, "CFS50_Installation_Folder", buff, sizeof(buff)) >= 0) {
                    p = buff + strlen(buff);
                    strcpy(p, "\\cfserver.exe");
                } else {
                    buff[0] = 0;
                }
                c = get_phclsid_from_file(buff);
                if(c) {
                    strncpy(clsidbuff, c, BUFFSZ);
                    clsidbuff[BUFFSZ] = 0;
                } else {
                    clsidbuff[0] = 0;
                    if(ret_clsid) return(-1);
                }
            }
        }
    }
    if(ret_clsid) {
        strcpy(ret_clsid, clsidbuff);
        return(0);
    }

    tot = 0;
    if(RegOpenKeyEx(HKEY_CURRENT_USER, KEYPATH "\\Client", 0, KEY_READ, &key)) return(tot);
    p = regbuff + sprintf(regbuff, "%s\\", KEYPATH "\\Client");

    for(k = 0; ; k++) {
        len = BUFFSZ;
        if(RegEnumKeyEx(key, k, p, (void *)&len, NULL, NULL, NULL, NULL)) break;
        strcat(p, "\\ProfileInfo");

        if(!regkey(HKEY_CURRENT_USER, regbuff, "ActivationData", buff, sizeof(buff))) {
            if(!buff[0]) continue;
            if(!clsidbuff[0]) {
                printf("- ActivationData found but no PhDeviceCLSID, so I can't decrypt it\n");
                continue;
            }
            printf("  ActivationData: %s\n\n", camfrog_pwd(buff, clsidbuff));
            tot++;
        }
        if(!regkey(HKEY_CURRENT_USER, regbuff, "Serial", buff, sizeof(buff))) {
            if(!buff[0]) continue;
            printf("  Serial:         %s\n\n", camfrog_pwd(buff, NULL));
            tot++;
        }
        if(!regkey(HKEY_CURRENT_USER, regbuff, "Nickname", buff, sizeof(buff))) {
            if(!buff[0]) continue;
            printf("  nickname:       %s\n", buff);
            if(regkey(HKEY_CURRENT_USER, regbuff, "Password", buff, sizeof(buff)) < 0) continue;
            printf("  password:       %s\n\n", camfrog_pwd(buff, NULL));
            tot++;
        }
    }

    printf("- \"Nickname List\\Item\" scanning:\n\n");

    for(k = 0; ; k++) {
        sprintf(regbuff, "%d", k);

        if(regkey(
            HKEY_CURRENT_USER,
            KEYPATH "\\Client" "\\CurrentVersion\\ProfileInfo\\Nickname List\\Item", regbuff,
            buff, sizeof(buff)) < 0) break;

        switch(k % 3) {
            case 0: printf("  nickname:    %s\n",   buff);                      break;
            case 1: printf("  password:    %s\n",   camfrog_pwd(buff, NULL));   break;
            case 2: printf("  ???          %s\n\n", camfrog_pwd(buff, NULL));   break;
        }
        tot++;
    }

    RegCloseKey(key);
    return(tot);
}
#endif



void show_pwd(u8 *msg, u8 *pwd, u8 *clsid) {
    printf("  %-22s %s\n", msg, camfrog_pwd(pwd, clsid));
}



// easy lame way, not correct but works
void camfrog_file(FILE *fd, u8 *clsid) {
    int     type;
    u8      buff[BUFFSZ + 1],
            *p,
            *s;

    for(;;) {
        type = fgetc(fd);
        if(type < 0) break;
        if(type > ' ') break;
    }
    fseek(fd, 0, SEEK_SET);
    while(fgets(buff, BUFFSZ, fd)) {
        for(p = buff; *p && (*p != '\n') && (*p != '\r'); p++);
        *p = 0;
        s = buff;

        if(type == '<') {
            p = stristr(s, "<ActivationData>");
            if(!p) p = stristr(s, "password>");
            if(p) {
                p = strchr(p, '>');
                if(!p) break;
                *p++ = 0;
                s = strchr(s, '<');
                if(!s) break;
                *s++ = 0;
                show_pwd(s, p, clsid);
            }
        }

        while(*s) {
            p = stristr(s, "Password=");
            if(!p) p = stristr(s, "param=");
            if(!p) break;
            for(s = p; s > buff; s--) { // not >=
                if(s[-1] <= ' ') break;
            }
            p = strchr(p, '=');
            if(!p) break;
            *p++ = 0;
            if(*p == '\"') p++;
            if(!stricmp(s, "remote_control_password")) continue;

            show_pwd(s, p, clsid);
            for(s = p; *s; s++) {
                if(strchr(" \t\r\n\"", *s)) break;
            }
            *s++ = 0;
        }
    }
}



int hex2byte(u8 *hex) {
    static const signed char hextable[256] =
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\xff\xff\xff\xff\xff\xff"
        "\xff\x0a\x0b\x0c\x0d\x0e\x0f\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\x0a\x0b\x0c\x0d\x0e\x0f\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff"
        "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff";

    if((hextable[hex[0]] < 0) || (hextable[hex[1]] < 0)) return(-1);
    return((hextable[hex[0]] << 4) | hextable[hex[1]]);
}



u8 *camfrog_pwd(u8 *input_pwd, u8 *clsid) {
    AES_KEY aes_ctx;
    int     i,
            t,
            pwdlen,
            clsidlen;
    static u8   *pwd = NULL;
    static const u8 key[] = "CamFrogMCU DLL01";

    pwdlen = strlen(input_pwd);
    pwd = realloc(pwd, pwdlen + 1);
    strcpy(pwd, input_pwd);

    for(i = 0; pwd[i << 1]; i++) {
        t = hex2byte(pwd + (i << 1));
        if(t < 0) break;
        pwd[i] = t;
    }
    pwdlen = i;
    if(!pwdlen) return(pwd);

    if((!pwd[1] && !pwd[2] && !pwd[3]) || (pwdlen % 16)) { // try to auto-guess
        strcpy(pwd, input_pwd);
        return(camfrogbot_pwd(pwd, clsid));
    }

    if(clsid) {
        for(i = 0; clsid[i << 1]; i++) {
            t = hex2byte(clsid + (i << 1));
            if(t < 0) break;
            clsid[i] = t;
        }
        clsidlen = i;

        for(i = 0; i < pwdlen; i++) {
            pwd[i] ^= clsid[i % clsidlen];
        }
    }

    if(AES_set_decrypt_key(key, strlen(key) << 3, &aes_ctx)) goto quit_error;
    for(i = 0; i < pwdlen; i += 16) {
        AES_decrypt(pwd + i, pwd + i, &aes_ctx);
    }
    pwd[i] = 0;
    return(pwd);
quit_error:
    return("missing or invalid PhDeviceCLSID, I can't decrypt");
}



u8 *camfrogbot_pwd(u8 *pwd, u8 *clsid) {
    AES_KEY aes_ctx;
    int     i,
            t,
            ivoff,
            pwdlen,
            real_pwdlen;
    u8      key[BUFFSZ + 1],    // "Internal Value"
            iv[16];

    if(!clsid) {
        if(camfrog_enum(key) < 0) goto quit_error;
    } else {
        t = strlen(clsid);
        if(t > BUFFSZ) goto quit_error;
        strcpy(key, clsid);
    }
    for(i = 0; (t = key[i]); i++) {
        if((t >= 'a') && (t <= 'z')) key[i] = toupper(t);
    }
    key[i & (~15)] = 0;
    for(i = 0; pwd[i << 1]; i++) {
        t = hex2byte(pwd + (i << 1));
        if(t < 0) break;
        pwd[i] = t;
    }
    pwdlen = i;
    if(!pwdlen) return(pwd);

    real_pwdlen = pwd[0] | (pwd[1] << 8) | (pwd[2] << 16) | (pwd[3] << 24);
    pwdlen -= 4;
    if((real_pwdlen < 0) || (real_pwdlen > pwdlen)) real_pwdlen = pwdlen;

    if(AES_set_encrypt_key(key, strlen(key) << 3, &aes_ctx)) goto quit_error;
    ivoff = 0;
    memset(iv, 0, sizeof(iv));
    AES_cfb128_encrypt(pwd + 4, pwd, pwdlen, &aes_ctx, iv, &ivoff, AES_DECRYPT);
    pwd[real_pwdlen] = 0;
    return(pwd);
quit_error:
    return("missing or invalid PhDeviceCLSID, I can't decrypt");
}


